//
// GameX - WindowsDX Class Header 
// 
// Copyright (C) 2002 Rama C. Hoetzlein
// 
// This software is released under the GameX GNU GPL 
// Open Source Liscense. See the GameX documentation included
// with this source code for terms of modification, 
// distribution and re-release. 
// 
#ifndef WINDX_DEF
	#define WINDX_DEF

	// Windows and other standard headers
	#include <windows.h>
	#include <stdio.h>
	#include <stdarg.h>
	#include <time.h>

	// DirectDraw
	#define DIRECTDRAW_VERSION 0x0700 // specify version 7.0 of DirectDraw
	#include <ddraw.h> // project must have DDRAW.LIB included too

	// Direct3D
	#define DIRECT3D_VERSION 0x0700 // specify version 7.0 of Direct3D -- the last version that's compatible with DirectDraw
	#include <d3d.h>				// note: for newer versions of D3D we would need to 
	#include <d3dtypes.h>			//		 include different files, like d3d8.h

	// DirectInput
	#define DIRECTINPUT_VERSION 0x0700 // specify version 7.0 of DirectInput
	#include <dinput.h> // project must have DINPUT.LIB included too

	// DirectSound
	#define DIRECTSOUND_VERSION 0x0700 // specify version 7.0 of DirectSound
	#include <dsound.h> // project must have DSOUND.LIB included too
	
	// DirectMusic
	#include <dmksctrl.h> // DirectMusic is a little different from the above includes,
	#include <dmusici.h>  // it's all COM so there is no such thing as DMUSIC.LIB
	#include <dmusicc.h>  // however, DXGUID.LIB must be included for COM stuff to link
	#include <dmusicf.h>

	// headers of other components of GameX we need
	#include "gamex-debug.hpp"
	#include "gamex-image.hpp"
	#include "gamex-sound.hpp"
	#include "gamex-vector.hpp"
	
	#ifdef _DEBUG
		#define WINDX_DEBUG						// Turns on WindowsDX debugging	
	#endif


	// combinable flags to be used with GameX.Initialize()
	// or later set with GameX.FlagOn() and GameX.FlagOff()
	// video flags:
	#define VIDEO_HIGHCOLOR				VIDEO_16BIT // currently the fastest and most common color mode, can show thousands of different colors
	#define VIDEO_TRUECOLOR				VIDEO_32BIT // often slower than VIDEO_HIGHCOLOR but can show millions of different colors (note: this is 32 bit if the monitor supports it, 24 bit if not, 16 bit if it supports neither)
	#define VIDEO_FULLSCREEN			0x00000001l	// run in fullscreen mode (at a resolution that best fits the requested viewport size)
	#define VIDEO_WINDOWED				0x00000002l	// run in windowed mode -- if neither VIDEO_FULLSCREEN nor VIDEO_WINDOWED is specified, the user will be prompted as to whether to run in fullscreen mode when the game starts
	#define VIDEO_LETTERBOX				0x00000004l	// if in fullscreen and if the requested size can't be matched exactly, disallow the game from stretching fully instead of only as proportionality allows
	#define VIDEO_NORESIZE				0x00000008l	// disallow user from resizing the game window if in windowed mode -- not recommended, there's no reason a game would need this since GameX automatically responds to resizes
	// other flags:
	#define RUN_KEEPCURSOR				0x10000000l	// prevent GameX from hiding the cursor when it is over the game window -- note that a visible cursor over the game window can reduce the game's framerate considerably, it's better to draw your own cursor in the game if you need it
	#define RUN_NOCONFIRMQUIT			0x20000000l	// disallow GameX from asking whether the user is sure about quitting
	#define RUN_NOMOUSEINPUT			0x40000000l	// prevent GameX from keeping track of mouse movement and clicks
	// video flags for advanced users:
	#define VIDEO_INDEXEDCOLOR			VIDEO_8BIT  // (NOT YET IMPLEMENTED) somewhat outdated color mode, uses a palette of only 256 indexed (but changeable) colors
	#define VIDEO_NOREFRESHSYNC			0x00000010l	// don't sync screen updates to monitor refresh rate -- allows framerate to exceed monitor refresh rate, but may cause screen chopping. Can use more or less power depending on frame rate.
	#define VIDEO_ALLOWPAGEFLIP			0x00000020l	// tells GameX to utilize page flipping when in fullscreen mode -- has no effect unless requested window size exactly matches a monitor resolution, such as 640x480 or 800x600 -- page flipping *can* lead to smoother animation, but can also cause flickering problems, dialog box disfunction, and much higher power usage
	#define VIDEO_TRIPLEBUFFER			0x00000040l	// (NOT YET IMPLEMENTED) attempt to use triple buffering for animation, must be used in conjunction with VIDEO_ALLOWPAGEFLIP
	#define VIDEO_UPDATEBETWEENSTATES	0x00000080l	// (NOT YET IMPLEMENTED) if the frame rate exceeds the cycle rate, allow GameX to update the extra frames between calls to main() -- only use if your draw() can draw the same frame differently, for instance if it uses randomness when drawing, and if you want that difference to be visible between game state updates
	#define VIDEO_DRAWALLSTATES			0x00000100l	// (NOT YET IMPLEMENTED) call draw() once after every call to main(), does not change number of screen updates -- not recommended, performance requirements can go way up, easily making gameplay choppier and erratic
	#define VIDEO_LOWQUALITY			0x00000200l	// use this if you want to see how the game will appear in low-quality mode, which normally you're only given the option to switch to if GameX discovers lack of hardware support during startup or if the game is running really slowly. This setting does not attempt to simulate the low framerate that lack of hardware support usually generates, if you want to test that then go to Control Panels -> Display Settings -> Advanced -> Troubleshoot -> Hardware Acceleration -> None
	#define VIDEO_SOFTWAREONLY			0x00000400l	// use this to force all D3D operations to use software instead of any available hardware -- usually a bad idea, unless using an outdated video card which doesn't fully support Direct3D
	#define VIDEO_8BIT					0x00001000l	// (NOT YET IMPLEMENTED) run game in 8 bit indexed color mode -- not implemented because of palette complications and the fact that 8-bit mode is outdated
	#define VIDEO_16BIT					0x00002000l	// run game in 16 bit color mode -- supported
	#define VIDEO_24BIT					0x00004000l	// (NOT YET TESTED) run game in 24 bit color mode -- supported -- externally it's the same thing as 32 bit color mode -- untested since I don't have a monitor that supports 24-bit color, only 32-bit, 16-bit, and 8-bit
	#define VIDEO_32BIT					0x00008000l	// run game in 32 bit color mode -- supported but may be slow depending on video card
	// audio flags for advanced users:
	#define AUDIO_DISABLE				0x00010000l // disable all audio support, so sounds and music will not play -- only possible use is to mute a game when debugging to make debugging easier or something
	#define AUDIO_NOSTEREO				0x00020000l // ignore sound panning arguments
	#define AUDIO_REVERSESPEAKERS		0x00040000l // reverse sound panning
	// other flags for advanced users:
	#define RUN_BACKGROUND				0x01000000l	// allow game to keep running even in the background by preventing GameX from automatically pausing the game -- possible uses are for networked games or cheat prevention, maybe
	#define RUN_UNMANAGED				0x02000000l	// prevent GameX from managing the game's speed and frame rate at all (just main() then draw() repeatedly), in which case the speed will be limited to the refresh rate unless VIDEO_NOREFRESHSYNC is on -- highly unrecommended, RUN_UNMANAGED should only be used in conjunction with custom game speed and frame rate management code (with Sleep() used to prevent straining the CPU too much)
	#define RUN_ASSUMEVRAMSAFE			0x04000000l	// (NOT YET TESTED) allow GameX to delete image data from RAM that could be kept solely in VRAM as a surface

	#define GAMEX_NO_INIT				0
	#define GAMEX_INITCLASS_FAILED		1
	#define GAMEX_INITCLASS_OK			2
	#define GAMEX_INITWINTEMP_FAILED	3
	#define GAMEX_INITWINTEMP_OK		4
	#define GAMEX_INITWINREAL_FAILED	5
	#define GAMEX_INITWINREAL_OK		6
	#define GAMEX_SHOWWIN_FAILED		7
	#define GAMEX_SHOWWIN_OK			8
	#define GAMEX_DDSTART_FAILED		9
	#define GAMEX_DDSTART_OK			10
	#define GAMEX_DDSETLEVEL_FAILED		11
	#define GAMEX_DDSETLEVEL_OK			12
	#define GAMEX_SETMODE_FAILED		13
	#define GAMEX_SETMODE_OK			14
	#define GAMEX_FRONTSURFACE_FAILED	15
	#define GAMEX_FRONTSURFACE_OK		16
	#define GAMEX_BACKSURFACE_FAILED	17
	#define GAMEX_BACKSURFACE_OK		18	
	#define GAMEX_READY					19
	#define GAMEX_LOCK_FAILED			20
	#define GAMEX_REQSIZE_FAILED		21
	#define GAMEX_BPP_FAILED			22
	#define GAMEX_SOUNDINIT_FAILED		23
	#define GAMEX_SOUNDCOOP_FAILED		24
	#define GAMEX_SOUNDCREATE_FAILED	25
	#define GAMEX_SOUNDMATCH_FAILED		26
	#define GAMEX_SOUNDCAPS_FAILED		27
	#define GAMEX_SOUNDBUFLEN_FAILED	28
	#define GAMEX_SOUNDLOST_FAIL		29
	#define GAMEX_SOUNDINVALID_FAIL		30
	#define GAMEX_SOUNDPARAM_FAIL		31
	#define GAMEX_SOUNDPRIOLEVEL_FAIL	32
	#define GAMEX_IMAGEUSE_FAILED		33
	#define GAMEX_CLIPPER_FAILED		34
	#define GAMEX_SHUTTINGDOWN			35
	#define GAMEX_D3DSTART_FAILED		36

	// input byte codes -- only to be used internally
	#define INPUT_DOUBLE	0x01 // custom bit code for double-pressed (for mouse double-click recognition)
	#define INPUT_HELD		0x20 // custom bit code for pressed and held, if turned on, should also turn off INPUT_PRESSED
	#define INPUT_PRESSED	0x80 // (must be 0x80, it's part of DirectInput functions for pressed states)

	#define WINDX_BUFSIZE	1000 // Keyboard Buffer Size, newest data is lost if buffer overflows

	#define MOUSE_POS_RATE	1.0

	// mouse button constants for GetMousePress() -- these values must match the fields of di_mousestate.rgbButtons[]
	#define MOUSE_LEFT		0 // left click
	#define MOUSE_RIGHT		1 // right click
	#define MOUSE_MIDDLE	2 // middle click (clicking on the scroll wheel, which not all mouses have)
	#define MOUSE_MID		MOUSE_MIDDLE // abbreviation
	#define MOUSE_EXTRA		3 // no idea what the fourth button on a mouse is, but might as well support it, if it exists

	#define VIEWPORT	(ImageX *)0 // dummy image that's really the viewport -- there's probably a better way of doing this involving polymorphism, but this works fine and requires fewer changes
									// when you want to draw to or from the viewport, use this constant as the destination or source

#ifndef DRAW_MODES_DEFINED
	#define DRAW_MODES_DEFINED
	// Draw Modes: (not to be combined with each other)
	// Exactly one Draw Mode must be specified whenever drawing an image.
	#define DRAW_NORMAL			0x0001	// backward compatibility
	#define DRAW_ALPHA			0x0001  // backward compatibility - user can do LOAD_ALPHA.. then DRAW_ALPHA
	#define DRAW_MASKED			0x0001  // backward compatibility - user can do LOAD_MASKED.. then DRAW_MASKED
	#define DRAW_PLAIN			0x0001 // DEST = SOURCE -- draw image normally -- can still be combined with options and color/alpha blending (makes use of alpha layer if image has one)
	#define DRAW_ADD			0x0002 // DEST = DEST + SOURCE -- draw additively using r,g,b intensity if given (multiplies by alpha layer if image has one, although alpha layer is NOT necessary for ADDED images and may slow things down) (alpha arg < 255 scales down intensity)
	#define DRAW_ADDED			0x0002 // backward compatibility
	#define DRAW_SUBTRACT		0x0004 // DEST = DEST - SOURCE -- draw subtractively using r,g,b intensity if given (alpha arg < 255 scales down intensity) -- not supported on older video cards
	#define DRAW_GHOST			0x0008 // DEST = DEST*(1-SOURCE) + SOURCE*SOURCE -- draw blended using color-valued alpha and r,g,b values if given (a < 255 scales down intensity/opacity) i.e. a sort of auto-alpha effect
	#define DRAW_INVERT			0x0010 // DEST = DEST + SOURCE - 2*SOURCE*DEST -- draw by inverting destination proportionally to source colors (alpha < 255 scales down intensity) -- if you want to invert the source image when drawing, that's what 
	#define DRAW_BURN			0x0020 // DEST = SOURCE - SOURCE*SOURCE - DEST*SOURCE -- draw by inverting destination proportionally to greyness of source colors -- may not be supported on older video cards
	#define DRAW_INTENSIFY		0x0040 // DEST = DEST + 2*SOURCE*DEST -- draw additively, but add less as the destination becomes darker -- not supported on older video cards
	#define DRAW_MULTIPLY		0x0080 // DEST = DEST * SOURCE -- multiply destination by source -- probably not useful in most situations -- not supported on older video cards
	#define DRAWMODESMASK		0x00FF // DO NOT SPECIFY THIS FLAG -- this is a mask for all of the draw modes and is only to be used internally
	// Looking for DRAW_ALPHA or DRAW_MASKED? Load the image with LOAD_ALPHA or LOAD_MASKED and draw with DRAW_PLAIN to achieve these effects
	// Note: All images should have a black background color for the above modes to draw properly, except for DRAW_MULTIPLY

	// Draw Options: (to be combined with a Draw Mode or also with each other)
	// Use these flags when drawing an image with DrawImage() or a similar function
	#define DRAWOP_BRIGHT		0x0100 // draw image with 2X brightness, so r,g,b of 128 act like 255, and r,g,b of 255 act like 510 (which is normally out of range) -- try combining with various draw modes for interesting effects
	#define DRAWOP_INVERTED		0x0200 // temporarily invert the image before drawing it -- try combining with various draw modes for interesting effects
	#define DRAWOP_NOFILTER		0x0400 // render the image without doing bilinear filtering, i.e. turns off automatic image smoothing
									   // avoid using this flag as much as possible, it will probably NOT increase speed, and only rarely will it make an image look better
	#define DRAWOP_DITHER		0x0800 // dither to give the impression that there are more colors than really available (usually with diffusion-type pixel dithering),
									   // i.e. turns on automatic color smoothing -- works best on gradual color changes, like gradients (i.e. DrawBox()), or on very faint image drawing
									   // only affects colors that can't be drawn accurately in the current display mode, and won't restore color accuracy that may have been lost upon loading an image with higher depth than the display mode
	#define DRAWOP_CENTERED		0x1000 // makes point coordinates specify center instead of top-left
									   // useful for many objects whose positions are defined in relation to their center
	#define DRAWOP_HMIRROR		0x2000 // horizontally flip the image when drawing -- the flip happens before any scaling/rotation/vertex blending
	#define DRAWOP_VMIRROR		0x4000 // vertically flip the image when drawing -- the flip happens before any scaling/rotation/vertex blending
	#define DRAWOP_SCALECANFLIP	0x8000 // if a scale is specified and the scale (x, y, or both) is negative, normally the image is not drawn,
									   // but with this flag the image will instead be drawn as if the scale is positive
									   // except it will also be mirrored about the axis of the scaling that was negative
	#define DRAWOP_ALTERNABLE 0x10000l // if in low quality mode (to correct low framerate), allow GameX to skip this drawing every other frame
	#define DRAWOP_SKIPPABLE  0x20000l // if in low quality mode (to correct low framerate), allow GameX to skip this drawing completely
	#define DRAWOP_PLAINABLE  0x40000l // if in low quality mode (to correct low framerate), allow GameX to draw with DRAW_PLAIN instead of whatever other mode is specified, and also without dithering/filtering/shading
	#define DRAWOPTIONSMASK	  0x7FF00l // DO NOT SPECIFY THIS FLAG -- to be used internally
#endif


	#define DEFAULT		-99999 // reserved value that signifies either a default color or image rectangle full expansion (or rarely, non-scaled centered changed to DEFAULT2)
	#define DEFAULT2	-99998 // reserved value that signifies image rectangle non-scaled expansion

	#define ENTIRE		DEFAULT // value


	class WindowsDX {
	public:
		WindowsDX (void);
		~WindowsDX (void);

		void SetGameName (char* name)	{strcpy (game_name, name); UpdateCaption();}
		void UpdateCaption ();

		//**** Initialize GameX
		// minimum necessary: call GameX.Initialize(options, xres, yres)
		// standard functionality: GameX.Initialize(name, options, xres, yres, cps)  (cps = speed)
		// do not specify fps or max_cpf except for games with very specific/unusual requirements
		bool Initialize (char * name, int options, int xres, int yres, int cps = 60, int fps = DEFAULT, int max_cpf = DEFAULT);
		inline bool Initialize (int options, int xres, int yres, int cps = 60, int fps = DEFAULT, int max_cpf = DEFAULT) {return Initialize("GameX Game",options,xres,yres,cps,fps,max_cpf);}

		// asks the user if they want to quit (unless RUN_NOCONFIRMQUIT is part of the Initialize options)
		// and unless the user says they don't want to quit,
		// quits the game completely and returns to Windows
		void Quit (void);

		// change things that were set when Initialize() was called,
		// if you have to change them while the game is still running:
		void ChangeGameSpeed (int cps) {win_cps = cps;}
//		void ChangeGameSize (int xres, int yres); // not yet implemented
//		void FlagOn (int flags); // not yet implemented
//		void FlagOff (int flags); // not yet implemented
		
		//**** Interface Functions
		void SetView (int x1, int y1, int x2, int y2);			// Set viewport		
		bool UpdateInput (void);								// Retrieve input

		bool DrawOpen (void);									// Open graphics output (maybe should make private)
		inline void PrepareToDrawPixels (void) {DrawOpen();} // prepares to draw pixels (only used because individual unlock checks for each pixel cause slowdown)
		inline void DrawClose (void) { if(drawLocked) {dd_back_buf->Unlock (NULL) ; drawLocked = false;} }	// unlock drawing surface

		void DrawColor (int r, int g, int b);					// Select drawing color		
		void GetDrawColor (int & r, int & g, int & b);

		inline void DrawPixel (int x,int y,int r,int g,int b)	// Draw pixel (use PrepareToDrawPixels() first!)
			{(*FuncDrawPixel) (x,y,r,g,b);}
		inline void DrawPixelAdded (int x,int y, int ri, int gi, int bi)	// Draw additive pixel (use PrepareToDrawPixels() first!)
			{(*FuncDrawPixelAdded) (x,y,ri,gi,bi);}


		// example of drawing an image:
		// GameX.DrawImage(DRAW_PLAIN, &image, x,y); // draws an image normally with its top-left at (x,y) onscreen
		// another example:
		// GameX.DrawImage(DRAW_ADD|DRAWOP_CENTERED, &photon, x,y, VIEWPORT, 128,255,128,255); // draws a photon additively and tinted blue, centered at (x,y) onscreen.
		// there are lots of complicated combinations available, so feel free to experiment.

		// master Draw function, used internally by GameX to enable DrawImage type functions to work
		// (should change this to accept 8 values instead of 4 for destination positions, to allow effects like skewing, perspective, etc.)
	       void DrawImageMaster(int flags, int x1s,int y1s,int x2s,int y2s, ImageX * source, int x1d,int y1d,int x2d,int y2d, ImageX * destination, int r,int g,int b,int a, int r2,int g2,int b2,int a2, int r3,int g3,int b3,int a3, int r4,int g4,int b4,int a4, float angle, float scalex, float scaley);
		
		// for the following draw functions:
		// - "flags" is used to determine how to draw the function, valid flag parameters include things like (DRAW_ADD) or (DRAW_PLAIN|DRAWOP_BRIGHT).
		// - source is a pointer to the ImageX you're drawing, specify VIEWPORT instead if you want to draw FROM the screen.
		// - x,y pairs form a source or destination point used to determine drawing location. x1,y1,x2,y2 pairs determine a rectangle.
		// - destination is a pointer to the ImageX you're drawing INTO, specify VIEWPORT to draw to the screen (which is what you'll usually want to do)
		//   or you can leave the destination parameter out if there are no parameters after it, and VIEWPORT will be assumed
		// - the optional parameter angle determines the amount to rotate the image when drawing it,
		//   the angle is in in degrees clockwise.
		// - the optional parameters scalex and scaley determine the amount to scale the image when drawing.
		//   if you only specify scalex, then scaley is assumed to equal scalex.
		//   the scale parameters are float values, 1.0f means 100%, 0.0f means 0% (nothing drawn), 0.5f means 50%, 1.5f means 150%, and if DRAWOP_SCALECANFLIP is on, -1.0f means 100% and flipped around
		//   if you specify a scale, the destination (well, the temporary copy of its boundary points) will be altered by the scaling amount.
		// - since the angle parameter comes before the scale parameter(s), you must specify an angle of 0.0f
		//   if you want to scale an image but not rotate it,
		//   or you can (of course) specify a non-zero angle and a scale to combine scaling and rotation
		//   but YOU MUST SPECIFY AN ANGLE (which can be zero) IN ORDER TO SPECIFY A SCALE.

		// draw an entire image, starting at a given point of the destination - with an angle
		inline void DrawImage(int flags,                                  ImageX * source, int xd,int yd, float angle, ImageX * destination=VIEWPORT,					 float scalex=DEFAULT, float scaley=DEFAULT)		{DrawImageMaster(flags, 0,0,DEFAULT,DEFAULT, source, xd,yd,DEFAULT2,DEFAULT2, destination, DEFAULT,DEFAULT,DEFAULT,DEFAULT, DEFAULT,DEFAULT,DEFAULT,DEFAULT, DEFAULT,DEFAULT,DEFAULT,DEFAULT, DEFAULT,DEFAULT,DEFAULT,DEFAULT, angle, scalex,scaley);};

		// draw an entire image, starting at a given point of the destination
		inline void DrawImage(int flags,                                  ImageX * source, int xd,int yd,                   ImageX * destination=VIEWPORT, float angle=0.0f, float scalex=DEFAULT, float scaley=DEFAULT)		{DrawImageMaster(flags, 0,0,DEFAULT,DEFAULT, source, xd,yd,DEFAULT2,DEFAULT2, destination, DEFAULT,DEFAULT,DEFAULT,DEFAULT, DEFAULT,DEFAULT,DEFAULT,DEFAULT, DEFAULT,DEFAULT,DEFAULT,DEFAULT, DEFAULT,DEFAULT,DEFAULT,DEFAULT, angle, scalex,scaley);};

		// draw a rectangular portion of an image, starting at a given point of the destination
		inline void DrawImage(int flags, int x1s,int y1s,int x2s,int y2s, ImageX * source, int xd,int yd,                   ImageX * destination=VIEWPORT, float angle=0.0f, float scalex=DEFAULT, float scaley=DEFAULT)		{DrawImageMaster(flags, x1s,y1s,x2s,y2s, source, xd,yd,DEFAULT2,DEFAULT2, destination, DEFAULT,DEFAULT,DEFAULT,DEFAULT, DEFAULT,DEFAULT,DEFAULT,DEFAULT, DEFAULT,DEFAULT,DEFAULT,DEFAULT, DEFAULT,DEFAULT,DEFAULT,DEFAULT, angle, scalex,scaley);};

		// draw an entire image into a rectangle of the destination
		inline void DrawImage(int flags,                                  ImageX * source, int x1d,int y1d,int x2d,int y2d, ImageX * destination=VIEWPORT, float angle=0.0f, float scalex=DEFAULT, float scaley=DEFAULT)		{DrawImageMaster(flags, 0,0,DEFAULT,DEFAULT, source, x1d,y1d,x2d,y2d, destination, DEFAULT,DEFAULT,DEFAULT,DEFAULT, DEFAULT,DEFAULT,DEFAULT,DEFAULT, DEFAULT,DEFAULT,DEFAULT,DEFAULT, DEFAULT,DEFAULT,DEFAULT,DEFAULT, angle, scalex,scaley);};

		// draw a rectangular portion of an image into a rectangle of the destination
		inline void DrawImage(int flags, int x1s,int y1s,int x2s,int y2s, ImageX * source, int x1d,int y1d,int x2d,int y2d, ImageX * destination=VIEWPORT, float angle=0.0f, float scalex=DEFAULT, float scaley=DEFAULT)		{DrawImageMaster(flags, x1s,y1s,x2s,y2s, source, x1d,y1d,x2d,y2d, destination, DEFAULT,DEFAULT,DEFAULT,DEFAULT, DEFAULT,DEFAULT,DEFAULT,DEFAULT, DEFAULT,DEFAULT,DEFAULT,DEFAULT, DEFAULT,DEFAULT,DEFAULT,DEFAULT, angle, scalex,scaley);};

		// draw part of an image starting at a point of the source, into a given destination rectangle
		inline void DrawImage(int flags, int xs, int ys,                  ImageX * source, int x1d,int y1d,int x2d,int y2d, ImageX * destination=VIEWPORT, float angle=0.0f, float scalex=DEFAULT, float scaley=DEFAULT)		{DrawImageMaster(flags, xs,ys,DEFAULT2,DEFAULT2, source, x1d,y1d,x2d,y2d, destination, DEFAULT,DEFAULT,DEFAULT,DEFAULT, DEFAULT,DEFAULT,DEFAULT,DEFAULT, DEFAULT,DEFAULT,DEFAULT,DEFAULT, DEFAULT,DEFAULT,DEFAULT,DEFAULT, angle, scalex,scaley);};

		// NOT YET IMPLEMENTED -- draw a rectangular portion of an image into a given destination, mapped onto a given quadrilateral
	//	inline void DrawImage(int flags, int x1s,int y1s,int x2s,int y2s, ImageX * source, int x1d,int y1d,int x2d,int y2d,int x3d,int y3d,int x4d,int y4d, ImageX * destination=VIEWPORT, float angle=0.0f, float scalex=DEFAULT, float scaley=DEFAULT)		{DrawImageMaster(flags, x1s,y1s,x2s,y2s, source, xd,yd,DEFAULT2,DEFAULT2, destination, DEFAULT,DEFAULT,DEFAULT,DEFAULT, DEFAULT,DEFAULT,DEFAULT,DEFAULT, DEFAULT,DEFAULT,DEFAULT,DEFAULT, DEFAULT,DEFAULT,DEFAULT,DEFAULT, angle, scalex,scaley);};

		// NOT YET IMPLEMENTED -- draw a rectangular portion of an image into a given destination, mapped onto a given polygon
	//	inline void DrawImage(int flags, int x1s,int y1s,int x2s,int y2s, ImageX * source, CONST POINT * points, int numpoints, ImageX * destination=VIEWPORT, float angle=0.0f, float scalex=DEFAULT, float scaley=DEFAULT);

		// the following draw functions are the same as the DrawImage ones above except:
		// - you can specify r,g,b,a for red, green, blue, alpha amounts to alter the coloring/transparency of the drawn image,
		//   or, you can specify 4 sets of r,g,b,a for each corner of the image, and they will be gauraud-shaded across it
		// - r,g,b,a are values that range from 0 to 255, 255 meaning full brightness or opacity
		// - if the DRAWOP_BRIGHT flag is specified, the red, green, and blue values you give are effectively doubled (internally)
		//   such that 128 means full brightness and 255 means double brightness -- the alpha is not changed by this, although many draw modes use brightness as alpha.

		// with r,g,b,a, shading for the entire image:
		inline void DrawImage(int flags,                                  ImageX * source, int xd, int yd,                  ImageX * destination, int r,int g,int b,int a,                                                                                        float angle=0.0f, float scalex=DEFAULT, float scaley=DEFAULT)		{DrawImageMaster(flags, 0,0,DEFAULT,DEFAULT, source, xd,yd,DEFAULT2,DEFAULT2, destination, r,g,b,a, DEFAULT,DEFAULT,DEFAULT,DEFAULT, DEFAULT,DEFAULT,DEFAULT,DEFAULT, DEFAULT,DEFAULT,DEFAULT,DEFAULT, angle, scalex,scaley);};
		inline void DrawImage(int flags, int x1s,int y1s,int x2s,int y2s, ImageX * source, int xd, int yd,                  ImageX * destination, int r,int g,int b,int a,                                                                                        float angle=0.0f, float scalex=DEFAULT, float scaley=DEFAULT)		{DrawImageMaster(flags, x1s,y1s,x2s,y2s, source, xd,yd,DEFAULT2,DEFAULT2, destination, r,g,b,a, DEFAULT,DEFAULT,DEFAULT,DEFAULT, DEFAULT,DEFAULT,DEFAULT,DEFAULT, DEFAULT,DEFAULT,DEFAULT,DEFAULT, angle, scalex,scaley);};
		inline void DrawImage(int flags,                                  ImageX * source, int x1d,int y1d,int x2d,int y2d, ImageX * destination, int r,int g,int b,int a,                                                                                        float angle=0.0f, float scalex=DEFAULT, float scaley=DEFAULT)		{DrawImageMaster(flags, 0,0,DEFAULT,DEFAULT, source, x1d,y1d,x2d,y2d, destination, r,g,b,a, DEFAULT,DEFAULT,DEFAULT,DEFAULT, DEFAULT,DEFAULT,DEFAULT,DEFAULT, DEFAULT,DEFAULT,DEFAULT,DEFAULT, angle, scalex,scaley);};
		inline void DrawImage(int flags, int x1s,int y1s,int x2s,int y2s, ImageX * source, int x1d,int y1d,int x2d,int y2d, ImageX * destination, int r,int g,int b,int a,                                                                                        float angle=0.0f, float scalex=DEFAULT, float scaley=DEFAULT)		{DrawImageMaster(flags, x1s,y1s,x2s,y2s, source, x1d,y1d,x2d,y2d, destination, r,g,b,a, DEFAULT,DEFAULT,DEFAULT,DEFAULT, DEFAULT,DEFAULT,DEFAULT,DEFAULT, DEFAULT,DEFAULT,DEFAULT,DEFAULT, angle, scalex,scaley);};
		inline void DrawImage(int flags, int xs, int ys,                  ImageX * source, int x1d,int y1d,int x2d,int y2d, ImageX * destination, int r,int g,int b,int a,                                                                                        float angle=0.0f, float scalex=DEFAULT, float scaley=DEFAULT)		{DrawImageMaster(flags, xs,ys,DEFAULT2,DEFAULT2, source, x1d,y1d,x2d,y2d, destination, r,g,b,a, DEFAULT,DEFAULT,DEFAULT,DEFAULT, DEFAULT,DEFAULT,DEFAULT,DEFAULT, DEFAULT,DEFAULT,DEFAULT,DEFAULT, angle, scalex,scaley);};
	//	inline void DrawImage(int flags, int x1s,int y1s,int x2s,int y2s, ImageX * source, int x1d,int y1d,int x2d,int y2d,int x3d,int y3d,int x4d,int y4d, ImageX * destination, int r,int g,int b,int a, float angle=0.0f, float scalex=DEFAULT, float scaley=DEFAULT)		{DrawImageMaster(flags, x1s,y1s,x2s,y2s, source, xd,yd,DEFAULT2,DEFAULT2, destination, r,g,b,a, DEFAULT,DEFAULT,DEFAULT,DEFAULT, DEFAULT,DEFAULT,DEFAULT,DEFAULT, DEFAULT,DEFAULT,DEFAULT,DEFAULT, angle, scalex,scaley);};
	//	inline void DrawImage(int flags, int x1s,int y1s,int x2s,int y2s, ImageX * source, CONST POINT * points, int numpoints, ImageX * destination, int r,int g,int b,int a, float angle=0.0f, float scalex=DEFAULT, float scaley=DEFAULT);

		// with different r,g,b,a, shading for each corner of the image:
		inline void DrawImage(int flags,                                  ImageX * source, int xd, int yd,                  ImageX * destination, int r,int g,int b,int a, int r2,int g2,int b2,int a2, int r3,int g3,int b3,int a3, int r4,int g4,int b4,int a4, float angle=0.0f, float scalex=DEFAULT, float scaley=DEFAULT)		{DrawImageMaster(flags, 0,0,DEFAULT,DEFAULT, source, xd,yd,DEFAULT2,DEFAULT2, destination, r,g,b,a, r2,g2,b2,a2, r3,g3,b3,a3, r4,g4,b4,a4, angle, scalex,scaley);};
		inline void DrawImage(int flags, int x1s,int y1s,int x2s,int y2s, ImageX * source, int xd, int yd,                  ImageX * destination, int r,int g,int b,int a, int r2,int g2,int b2,int a2, int r3,int g3,int b3,int a3, int r4,int g4,int b4,int a4, float angle=0.0f, float scalex=DEFAULT, float scaley=DEFAULT)		{DrawImageMaster(flags, x1s,y1s,x2s,y2s, source, xd,yd,DEFAULT2,DEFAULT2, destination, r,g,b,a, r2,g2,b2,a2, r3,g3,b3,a3, r4,g4,b4,a4, angle, scalex,scaley);};
		inline void DrawImage(int flags,                                  ImageX * source, int x1d,int y1d,int x2d,int y2d, ImageX * destination, int r,int g,int b,int a, int r2,int g2,int b2,int a2, int r3,int g3,int b3,int a3, int r4,int g4,int b4,int a4, float angle=0.0f, float scalex=DEFAULT, float scaley=DEFAULT)		{DrawImageMaster(flags, 0,0,DEFAULT,DEFAULT, source, x1d,y1d,x2d,y2d, destination, r,g,b,a, r2,g2,b2,a2, r3,g3,b3,a3, r4,g4,b4,a4, angle, scalex,scaley);};
		inline void DrawImage(int flags, int x1s,int y1s,int x2s,int y2s, ImageX * source, int x1d,int y1d,int x2d,int y2d, ImageX * destination, int r,int g,int b,int a, int r2,int g2,int b2,int a2, int r3,int g3,int b3,int a3, int r4,int g4,int b4,int a4, float angle=0.0f, float scalex=DEFAULT, float scaley=DEFAULT)		{DrawImageMaster(flags, x1s,y1s,x2s,y2s, source, x1d,y1d,x2d,y2d, destination, r,g,b,a, r2,g2,b2,a2, r3,g3,b3,a3, r4,g4,b4,a4, angle, scalex,scaley);};
		inline void DrawImage(int flags, int xs, int ys,                  ImageX * source, int x1d,int y1d,int x2d,int y2d, ImageX * destination, int r,int g,int b,int a, int r2,int g2,int b2,int a2, int r3,int g3,int b3,int a3, int r4,int g4,int b4,int a4, float angle=0.0f, float scalex=DEFAULT, float scaley=DEFAULT)		{DrawImageMaster(flags, xs,ys,DEFAULT2,DEFAULT2, source, x1d,y1d,x2d,y2d, destination, r,g,b,a, r2,g2,b2,a2, r3,g3,b3,a3, r4,g4,b4,a4, angle, scalex,scaley);};
	//	inline void DrawImage(int flags, int x1s,int y1s,int x2s,int y2s, ImageX * source, int x1d,int y1d,int x2d,int y2d,int x3d,int y3d,int x4d,int y4d, ImageX * destination, int r,int g,int b,int a, int r2,int g2,int b2,int a2, int r3,int g3,int b3,int a3, int r4,int g4,int b4,int a4, float angle=0.0f, float scalex=DEFAULT, float scaley=DEFAULT)		{DrawImageMaster(flags, x1s,y1s,x2s,y2s, source, xd,yd,DEFAULT2,DEFAULT2, destination, r,g,b,a, r2,g2,b2,a2, r3,g3,b3,a3, r4,g4,b4,a4, angle, scalex,scaley);};
	//	inline void DrawImage(int flags, int x1s,int y1s,int x2s,int y2s, ImageX * source, CONST POINT * points, int numpoints, ImageX * destination, const int * reds, const int * greens, const int * blues, const int * alphas, float angle=0.0f, float scalex=DEFAULT, float scaley=DEFAULT);

		// to help you remember parameter order for DrawImage functions:
		//
		// how to draw, where source, what source, where dest, what dest,   extra options
		// flags        xs,ys...      source       xd,yd...    destination  r,g,b,a,angle...
		// 
		//                   <HOWtoDRAW> <WHEREtoDRAWfrom---------------> <WHATtoDRAWfrom> <WHEREtoDRAWinto----------------> <WHATtoDRAWinto-----> <EXTRA-OPTIONS(COLOR)-----------------------------------------------------------------------------------------> <EXTRA OPTIONS(TRANSFORM)---------------------------------->







		// the full version of DrawBox (supports 3D-accelerated alpha blending and vertex shading)
		       void DrawBox(int flags, int x1, int y1, int x2, int y2, int r,int g,int b,int a, int r2,int g2,int b2,int a2, int r3,int g3,int b3,int a3, int r4,int g4,int b4,int a4, float angle = 0.0f);

		// functions with fewer parameters that are inlined to call the above function as appropriate:
		inline void DrawBox(int flags, int x1, int y1, int x2, int y2, int r,int g,int b,       int r2,int g2,int b2,        int r3,int g3,int b3,        int r4,int g4,int b4, int a, float angle = 0.0f) { DrawBox(flags, x1, y1, x2, y2, r, g, b, a, r2, g2, b2, a, r3, g3, b3, a, r4, g4, b4, a, angle); }
		inline void DrawBox(int flags, int x1, int y1, int x2, int y2, int r,int g,int b,int a,                                                                                        float angle = 0.0f) { DrawBox(flags, x1, y1, x2, y2, r, g, b, a, r, g, b, a, r, g, b, a, r, g, b, a, angle); }
		inline void DrawBox(int flags, int x1, int y1, int x2, int y2, int r,int g,int b,       int r2,int g2,int b2,        int r3,int g3,int b3,        int r4,int g4,int b4                           ) { DrawBox(flags, x1, y1, x2, y2, r, g, b, 255, r2, g2, b2, 255, r3, g3, b3, 255, r4, g4, b4, 255, 0.0f); }
		inline void DrawBox(int flags, int x1, int y1, int x2, int y2, int r,int g,int b                                                                                                                 ) { if(flags & DRAW_PLAIN) DrawBox(x1,y1,x2,y2,r,g,b); else DrawBox (flags,x1,y1,x2,y2,r,g,b,255,r,g,b,255,r,g,b,255,r,g,b,255,0.0f); }

		// the bare-bones (fast) version of DrawBox:
		       void DrawBox(int x1, int y1, int x2, int y2, int r, int g, int b);



		// draws an image tiled to fill the viewport
		// for average users who just want to draw a background,
		// simply call GameX.DrawBackground(&image);
		//          or GameX.DrawBackground(&image, x,y);
		//          or GameX.DrawBackground(&image, x,y, angle,scale); // to only use angle, set scale=1.0f, or to only use scale, set angle=0.0f
		// you can also give two sets of x,y, the first set (tx,ty) to be used for translation
		// and the second set (rx,ry) to be used (sort of indirectly) for rotational center,
		// but this is for advanced users only, since tx,ty,rx,ry takes a lot of fiddling to get right.
			   int DrawBackground (int flags, ImageX * img, int tx, int ty, int rx, int ry, int r,int g,int b,int a, float angle, float scale);
		inline int DrawBackground (int flags, ImageX * img, int  x, int  y,                 int r,int g,int b,int a, float angle, float scale) { return DrawBackground(flags, img,x,y,win_width/2,win_height/2, r,g,b,a, angle,scale); }
		inline int DrawBackground (           ImageX * img, int tx, int ty, int rx, int ry, int r,int g,int b,int a, float angle, float scale) { return DrawBackground(DRAW_PLAIN, img, tx, ty, rx, ry, r,g,b,a, angle, scale); }
		inline int DrawBackground (           ImageX * img, int  x, int  y,                 int r,int g,int b,int a, float angle, float scale) { return DrawBackground(DRAW_PLAIN,img,x,y,win_width/2,win_height/2, r,g,b,a, angle,scale); }
		inline int DrawBackground (int flags, ImageX * img, int tx, int ty, int rx, int ry,                          float angle, float scale) { return DrawBackground(flags, img, tx, ty, rx, ry, 255,255,255,255, angle, scale); }
		inline int DrawBackground (int flags, ImageX * img, int  x, int  y,                                          float angle, float scale) { return DrawBackground(flags, img,x,y,win_width/2,win_height/2, 255,255,255,255, angle,scale); }
		inline int DrawBackground (int flags, ImageX * img, int  x, int  y                                                                   ) { return DrawBackground(flags, img,x,y,0,0, 255,255,255,255, 0.0f,1.0f); }
		inline int DrawBackground (           ImageX * img, int tx, int ty, int rx, int ry,                          float angle, float scale) { return DrawBackground(DRAW_PLAIN, img, tx, ty, rx, ry, 255,255,255,255, angle, scale); }
		inline int DrawBackground (           ImageX * img, int  x, int  y,                                          float angle, float scale) { return DrawBackground(DRAW_PLAIN,img,x,y,win_width/2,win_height/2, 255,255,255,255, angle,scale); }
		inline int DrawBackground (           ImageX * img, int  x, int  y                                                                   ) { return DrawBackground(DRAW_PLAIN,img,x,y,0,0, 255,255,255,255, 0.0f,1.0f); }
		inline int DrawBackground (           ImageX * img                                                                                   ) { return DrawBackground(DRAW_PLAIN,img,0,0,0,0, 255,255,255,255, 0.0f,1.0f); }


	
		       void Blur(LPDIRECTDRAWSURFACE7 surface, RECT target_rect, float blur); // used internally for Blur

		// blurs a rectangle of (or all of) the target image or screen by a given amount (amount of 0.0f for no blue, 1.0f for total blur, 0.5f for decently large amount of blur)
		// note: currently may cause a blocky mosaic blur on old computers
		// note: these blur functions might take a little cleverness to use effectively
			   void Blur(ImageX * img, int x1, int y1, int x2, int y2, float amount);
		inline void Blur(              int x1, int y1, int x2, int y2, float amount)	{Blur(VIEWPORT,x1,y1,x2,y2,amount);}
		inline void Blur(                                              float amount)	{Blur(VIEWPORT,0,0,DEFAULT,DEFAULT,amount);}
		inline void Blur(ImageX * img,                                 float amount)	{Blur(img,0,0,DEFAULT,DEFAULT,amount);}


		// used to partially fade the entire viewport to a color determined by the viewport's average colors at each quadrant
		void SpecialViewFade(int a);

		// used by SpecialViewFade()
		void GetAverageViewportColors(int & r,int & g,int & b,int & r2,int & g2,int & b2,int & r3,int & g3,int & b3,int & r4,int & g4,int & b4);
		

		// Draws the given text at a point on the screen -- use DrawColor() beforehand to set color
		// note: this function LOCKS the drawing surface if it's unlocked,
		// whereas most image drawing functions UNLOCK it if it's locked.
		// therefore, alternating this function with drawing functions results in MAJOR slowdown
		int DrawText (int x, int y, char *msg);

		// same as DrawText, but draws a text outline of color (br,bg,bb) which are out of 255
		// is only a little slower than DrawText
		int DrawOutlinedText (int x, int y, int br, int bg, int bb, char *msg);

		// clear the view screen, setting all pixels to black
		void DrawClear (void);

		inline void DrawPolygon (CONST POINT * points, int num_points);


		// Updates the window with the view that's been drawn for this frame
		// The game should NOT call this function except under certain special circumstances
		// because this function is called automatically after each time draw() is called.
		// If you call this function in draw(), the view will wastefully update twice in a row.
		int ViewUpdate (void);

		// Sound functions for the game's code to call:

		// Plays a sound that has already been loaded into a SoundX object,
		// at specified volume (0.0f to 1.0f, 1.0f being full volume)
		// pan (-1.0f full left, 0.0f both speakers, 1.0f full right, 1.5f far right and half again volume)
		// and frequency multiplier (1.0f for normal, 2.0f for double speed and pitch, 0.5f for half speed and pitch)
		void PlaySound (SoundXPtr snd, int how=0, float vol=1.0f, float pan=0.0f, float freq=1.0f);
		void StopSound (SoundX* snd);
		void RewindSound (SoundX* snd);
		bool IsSoundPlaying (SoundXPtr snd);


		// Music functions:
		// (must load the music with music.LoadMidi("musicfile.mid"); before using the following)

		// Plays a MusicX object, and immediately stops playing any other music that may be playing
		// times of 0 means infinite loop, 1 means play once, 2 means play twice, etc.
		// speed_factor of 1.0 means normal speed, 2.0 means double speed, etc.
		void PlayMusic (MusicX * music, int times = 0, float fade_in_seconds = 1.0f, float volume = 1.0f, float tempo = 1.0f, float pitch = 1.0f);

		// Fades the volume to 0 over given seconds if seconds>0, then stops the music from playing
		void StopMusic (MusicX * music, float seconds = 2.0f);

		// Fades the music up or down to the given volume factor, over the given time
		// seconds is how many seconds to alter volume over
		// target_factor of 0.0f is fade to silence, 0.5f is fade to half master volume, 1.0f is no change, 1.27f is fade to max amplified
		void VolumeFadeMusic(MusicX * music, float volume_factor = 0.0f, float seconds = 0.0f);

		// Bends the pitch of the mucis up or down to the given pitch factor, over the given time
		// seconds is how many seconds to alter pitch over. This method does not change speed of play
		// pitch_factor of 0.5f is half pitch (low frequency), 2.0f is double pitch (high frequency), 1.0f is no change, etc.
		void PitchBendMusic(MusicX * music, float pitch_factor = 0.5f, float seconds = 0.0f);

		// Sets the tempo, or speed of play, of all music -- does not change the pitch
		// 0.5f is half speed, 1.0f is normal speed, 2.0f is double speed, etc.
		// Can be called before or after you start playing the song
		// note: there may be a small time delay after calling this function before the change takes effect
		void SetMusicTempo(float tempo_factor);

		// Returns whatever tempo was last set by GetMusicTempo, or the default of 1.0f
		float GetMusicTempo(void) {return master_music_tempo;}

		// Sets the volume of all music
		// Can be called before or after you start playing a song
		// If an individual music song has its volume faded to 50% and the master volume is 50%
		// that will result in the music being played at 25% volume
		void SetMasterMusicVolume(float volume_factor);

		// Returns whatever volume was last set by SetMasterMusicVolume, or the default of 1.0f
		float GetMasterMusicVolume(void) {return master_music_volume;}

		// returns true if the given song is still playing, looping, or paused, false otherwise
		bool IsMusicPlaying(MusicX * music);

		// pauses whatever MIDI song is currently playing (has only been tested internally to GameX)
		void PauseMusic (void);

		// resumes whatever MIDI song is currently playing (has only been tested internally to GameX)
		// due to the way it is paused, a stray note might play for a while after resuming
		void ResumeMusic (void);


		// Provided for backward compatibility. These functions are offered to those developers
		// who only need a simple interface, where the variety other input functions may be
		// too confusing. (Rama)
		inline bool GetKeyDown (int key_id)		{return IsKeyDown (key_id);}
		inline char GetKey ()					{return GetBufferedKeyPress();}

		// Keyboard Input Functions

		// returns the next char of buffered key input (like 'A' or 'b' or '3'), returns 0 if there is no key input left
		// use this to get typed data input from the user (such as a name at a name entry screen)
		inline char GetBufferedKeyPress (void)	{return (win_keys>0) ? win_keybuf[--win_keys] : 0;}

		// returns the next char of buffered key input without advancing past it, returns 0 if there is no key input left
		// use this if you want to see what GetBufferedKeyPress would return without actually taking it out of the buffer
		inline char PeekBufferedKeyPress (void)	{return (win_keys>0) ? win_keybuf[win_keys-1] : 0;}

		// completely clears the key buffer
		// use this, for instance, before asking for input to clear unwanted data from earlier
		inline void ClearKeyBuffer (void)		{win_keys = 0;}

		// for the following 3 functions:
		// valid key constants: DIK_LEFT, DIK_RIGHT, DIK_UP, DIK_DOWN, DIK_0 to DIK_9, DIK_A to DIK_Z, DIK_RETURN, DIK_SPACE, DIK_LALT, etc.
		// for all the key constants, open "dinput.h" and search for DIK_ESCAPE

		// returns true if the given key is down, false otherwise
		// use this if you only want to know if a given key is down or not (like the trigger of a machine gun)
		inline bool IsKeyDown (int key_id)		{return (di_keystate[key_id] & INPUT_PRESSED) ? true : false;}

 		// returns true if the given key is down and was up at least once since the last time this function was called on it, false otherwise
		// use this if you want to see if a key was just pressed (not just being held down, so you only respond once)
		inline bool GetKeyPress (int n)			{if (di_keypressed[n] & INPUT_PRESSED) {di_keypressed[n] ^= (INPUT_HELD|INPUT_PRESSED); return true;} else return false;}

		// returns true if the given key is down and was not held down since the last time GetKeyPress was called on it, false otherwise
		// use this if you want to see if a key was just pressed but don't want to also mark it as being held down yet
		inline bool PeekKeyPress (int n)		{return (di_keypressed[n] & INPUT_PRESSED) ? true : false;}
		
		// returns true if any key on the keyboard is down and was up at least once since GetKeyPress was called on it, false otherwise
		// use this at "Press Any Key" prompts. Note that this automatically excludes certain keys, such as volume control keys that a game should not respond to.
		bool AnyKeyPress (void);
		bool PeekAnyKeyPress (void);

		// convenience methods for checking for modifier keys being held down
		inline bool IsShiftDown (void)			{return IsKeyDown(DIK_LSHIFT)   || IsKeyDown(DIK_RSHIFT)  ;}
		inline bool IsAltDown (void)			{return IsKeyDown(DIK_LMENU)    || IsKeyDown(DIK_RMENU)   ;}
		inline bool IsCtrlDown (void)			{return IsKeyDown(DIK_LCONTROL) || IsKeyDown(DIK_RCONTROL);}



		// Mouse Functions

		inline int GetMouseX (void) {return mouse_x*win_width/win_realx;} // mouse x position, right is positive (scaled to the virtual screen so the game can use it without worrying about size variations that it can't control)
		inline int GetMouseY (void) {return mouse_y*win_height/win_realy;} // mouse y position, down is positive (scaled to the virtual screen)
		inline int GetMouseDX (void) {return di_mousestate.lX;} // mouse x movement, right is positive
		inline int GetMouseDY (void) {return di_mousestate.lY;} // mouse y movement, down is positive
		inline int GetMouseDZ (void) {return di_mousestate.lZ;} // mouse wheel movement, up is positive

		// sets the mouse position in terms of game pixels, (0,0) being top-left corner of the game view
		// valid (x,y) are between (0,0) and (win_width-1,win_height-1) 
		inline void SetMousePosition (int x, int y) {if(!di_mousestop) SetCursorPos(win_xpos+win_xoffset+x*win_realx/win_width,win_ypos+win_yoffset+y*win_realy/win_height);}

		// for the following 3 functions:
		// valid button constants: MOUSE_LEFT, MOUSE_MIDDLE, MOUSE_RIGHT
		
		// returns true if the given mouse button is down, false otherwise
		inline bool IsMouseDown(int button)		{return (di_mousestate.rgbButtons[button] & INPUT_PRESSED) ? true : false;}

		// returns true if the given mouse button is down and was up at least once since the last time this function was called on it, false otherwise
		inline bool GetMouseClick(int b)		{if(di_mousepressed[b] & INPUT_PRESSED) {di_mousepressed[b] ^= (INPUT_HELD|INPUT_PRESSED); return true;} else return false;}

		// returns true if the given mouse button is down and was not held down since the last time GetMouseClick was called on it, false otherwise
		// (i.e. returns true for as long as the mouse button is held down after a click, can be used for click-drags)
		inline bool PeekMouseClick(int b)		{return (di_mousepressed[b] & INPUT_PRESSED) ? true : false;}

		// returns true if any mouse button is down and was up at least once since GetMouseClick was called on it, false otherwise
		inline bool AnyMouseClick(void)			{for(int b=0; b<4; b++) if(GetMouseClick(b)) return true; return false;}
		inline bool PeekAnyMouseClick(void)		{for(int b=0; b<4; b++) if(PeekMouseClick(b)) return true; return false;}

		// returns true if the given mouse button is double-clicked and was up at least once since the last time this function was called on it, false otherwise
		// use this if you want to respond once to a double-click on a certain mouse button
		inline bool GetMouseDoubleClick(int b)	{if(di_mousepressed[b] & INPUT_DOUBLE) {di_mousepressed[b] ^= (INPUT_HELD|INPUT_DOUBLE); return true;} else return false;}

		// returns true if the given mouse button is double-clicked and was not held down since the last time GetMouseDoubleClick was called on it, false otherwise
		// (i.e. returns true for as long as the mouse button is held down after a double-click, can be used for double-click-drags)
		inline bool PeekMouseDoubleClick(int b)	{return (di_mousepressed[b] & INPUT_DOUBLE) ? true : false;}



		// Get the width or height of the (virtual) game screen,
		// The game's code should call these functions
		// to figure out the extents of where it can draw.
		// Unless, of course, it always initializes the game at one xres,yres
		// in which case it could just use hard-coded width and height values that it remembers
		inline int GetWidth (void)				{return win_width;}
		inline int GetHeight (void)				{return win_height;}
//		inline int GetPitch (void)				{return win_pitch;} // Only to be used by very advanced users who are attaching a custom drawing operation to GameX externally
		inline int GetCurrentFrameRate (void)	{return f_rate;} // current frame rate, higher is better up to a certain point, although the game must be in fullscreen mode for optimal smoothness
		inline int GetCurrentCycleRate (void)	{return c_rate;} // current game speed, something's very wrong if this is much different from what GetTargetCycleRate() returns
		inline int GetCurrentIdlePercent (void)	{return p_usage;} // lower is BETTER, higher means the game is using more CPU cycles -- if you want fullscreen mode to consume low power, specify VIDEO_NOREFRESHSYNC
		inline int GetTargetFrameRate (void)	{return min(win_fps,win_cps);}
		inline int GetTargetCycleRate (void)	{return win_cps;}
		inline void* GetScreen (void)			{return win_screen;}
		inline HWND GetWindow (void)			{return win_hwnd;}	
		inline LPDIRECTDRAW7 GetDD (void)		{return dd_main;}
		inline LPDIRECTSOUND GetDS (void)		{return ds_main;}
		inline IDirectMusicPerformance * GetDMP (void)	{return dm_performance;}
		inline IDirectMusicLoader * GetDML (void)		{return dm_loader;}
		inline bool IsFullScreen (void)			{return win_fullscreen;}
		inline bool SupportsDepth (int bpp)		{return (bpp == 16 || bpp == 24 || bpp == 32);}
		inline bool CanAutoPause (void)			{return (!win_request_no_auto_pause);}	
		inline bool CanHideCursor (void)		{return (!win_request_no_hide_cursor);}	
		inline bool CanManageSpeed (void)		{return (!win_request_no_manage_speed);}	
		inline bool CanUseMouse (void)			{return (!win_request_no_mouse_input);}	
		inline bool CanDeleteMData (void)		{return (win_request_delete_mdata);}	
		inline bool CanDraw (void)				{return (win_status == GAMEX_READY && !dd_back_buf->IsLost());}	
		inline bool MustDrawAll (void)			{return win_request_draw_all_states;}
		inline bool MustUpdateAll (void)		{return win_request_update_between_states;}
		inline int WasLost (void)				{return dd_was_lost;}
		inline char* GetName (void)				{return win_name;}	
		inline LPDIRECTDRAWSURFACE7 GetBackSurf (void) {return dd_back_buf;}
		bool CheckQuit (void);
		void QualityCheck (void);

		void HideCursor(void); // game code should NEVER call this!
		void UnhideCursor(void); // game code should NEVER call this! // calling this ShowCursor would create a name conflict with the windows API

		inline int GetStatus (void)				{return win_status;} // to be used internally only
		inline void SetStatus (int status)		{win_status = status;} // to be used internally only

		int GetShift (int mask);
		int GetMax (int mask);
		LPDIRECTDRAWSURFACE7 CreateImageSurface (int width, int height, bool texture, int create_alpha, bool targetable);
		bool CreateDeviceForSurface (LPDIRECT3DDEVICE7 & device, LPDIRECTDRAWSURFACE7 & surface);
///		inline int AssignDeviceToSurface (LPDIRECT3DDEVICE7 & device, LPDIRECTDRAWSURFACE7 & surface);
		//*** Windows Control Functions
		// Report a Problem Initializing
		void ReportProblem (void);		
		void ReportProblem (int problem);
		// Initialization & Clean Up Functions		
		void WindowInitInstance (HINSTANCE hThisInst, int nWinMode);	
		bool InitWindows (void);
		bool InitDirectX (void);
		bool InitSoundX (void);
		void DirectDrawError (int status);
		void Direct3DError (int status);
		bool CleanUp (void);		
		// Internal State Control Functions
		void SetActivation (BOOL act);
		void SetPosition (int x, int y);		
		void SetProportional (void);
		void InitMouseMove (void);
		void SetMouseMove (int x, int y); // to be used internally only, the game programmer should use SetMousePosition to change the mouse position
//		void SetMouseStop (void) {SetMouseMove(mouse_x,mouse_y);}
		void SetMouseStop (int disable_length) {di_mousestop = disable_length;} // to be used internally only
///		inline void SetKeyDown (char ch)		{win_keydown[ch] = true;}
///		inline void SetKeyUp (char ch)			{win_keydown[ch] = false;}
///		inline void ClearPressed (void)			{for(int k = 0 ; k < 256 ; k++) SetKeyUp(k);
///												 for(int m = 0 ; m < 3 ; m++)	SetMouseRelease(m);}



		void AddNetVar (void * var_address, int var_size)
		{
			// not yet implemented
		}





		// semi-private input functions
		void AddBufferedKeyPress (char ch);
		inline void SetMouseRelease (int button)		{mouse_down[button] = 0;}
		inline void SetMousePress (int button)			{mouse_down[button] = 1;}
		inline void SetMouseDoubleClick (int button)	{mouse_down[button] = 2;}

		// used internally when a MusicX object is loaded
		inline void RegisterMusicX (MusicX * mxp) {dm_musicx[dm_musicx_index++] = mxp;}


		// Public Variables
		char			*win_screen;
		int				win_ytable[2080];
		int				win_height, win_width; // width/height of the virtual game screen
		int				win_pitch, win_adjustedpitch; // pitch and bpp-adjusted pitch of the virtual game screen
		int				win_realx, win_realy; // width/height of visible window/fullscreen area, and bpp
		int				win_screenx, win_screeny, win_screenbpp, win_screenrefresh; // width/height of visible window/fullscreen area, and bpp
		int				win_viewx1, win_viewy1;
		int				win_viewx2, win_viewy2;		
		BOOL			win_active;
		int				win_cps; // speed of game -- number of times main() is called per second
		int				win_fps; // smoothness of game -- number of times draw() is called per second
		int				win_max_cpf; // "jerkiness" limit -- max cycles that can be skipped without drawing -- set really high if you want no limit
		unsigned short	f_rate; // current frame rate so game can know how many fps it's getting
		unsigned short	c_rate; // current cycle rate so game can know how many cps it's getting
		int				p_usage; // power usage (as a percent)
		int				sleep_record;
		bool			win_alternator;
		bool manual_pause;
		bool auto_pause;
///		LPDIRECTDRAWSURFACE dd_back_buf;
		bool drawLocked;
		int						win_xpos, win_ypos;		
		int						win_xoffset, win_yoffset;		
	private:		
		//**** Private Functions		

		void SetGameXVer (char* name)	{strcpy (gamex_ver, name); UpdateCaption();}

		// Dispatched Drawing Functions		
		// Draw an image using Direct3D (hopefully 3D-accelerated)
		int DrawImage3D(LPDIRECT3DDEVICE7 device, LPDIRECTDRAWSURFACE7 image, RECT & src_rect, RECT & src_bounds, RECT & dst_rect, RECT & dst_bounds, int surf_width, int surf_height, int r, int g, int b, int a, int r2, int g2, int b2, int a2, int r3, int g3, int b3, int a3, int r4, int g4, int b4, int a4, float angle, int flags, int alphatype);

		// Draw an image using DirectDraw (hopefully 3D-accelerated)
		int DrawImage2D(LPDIRECTDRAWSURFACE7 p_src_surface, LPDIRECTDRAWSURFACE7 p_dst_surface, RECT & src_rect, RECT & dst_rect, int nomask);

		// Draw an image by copying the pixels ourselves (no possibility for acceleration)
		int DrawPlain (XBYTE * source_pix, XBYTE * dest_pix, RECT & src_rect, RECT & dst_rect, RECT & src_bounds, RECT & dst_bounds, int src_pitch, int dst_pitch, int nomask);
		int DrawBlended (XBYTE * source_pix, XBYTE * dest_pix, RECT & src_rect, RECT & dst_rect, RECT & src_bounds, RECT & dst_bounds, int src_pitch, int dst_pitch, int r, int g, int b, int a, int nomask);
		int DrawAlphaUnmerged (XBYTE * source_pix, XBYTE * src_alpha_pix, XBYTE * dest_pix, RECT & src_rect, RECT & dst_rect, RECT & src_bounds, RECT & dst_bounds, int src_pitch, int dst_pitch, int alphafactor);
		int DrawAlphaMerged (XBYTE * source_pix, XBYTE * dest_pix, RECT & src_rect, RECT & dst_rect, RECT & src_bounds, RECT & dst_bounds, int src_pitch, int dst_pitch, int alphafactor);
		int DrawAdded (XBYTE * source_pix, XBYTE * dest_pix, RECT & src_rect, RECT & dst_rect, RECT & src_bounds, RECT & dst_bounds, int src_pitch, int dst_pitch, int ri, int gi, int bi);
		int DrawAddedMerged (XBYTE * source_pix, XBYTE * dest_pix, RECT & src_rect, RECT & dst_rect, RECT & src_bounds, RECT & dst_bounds, int src_pitch, int dst_pitch, int ri, int gi, int bi);
		int DrawInvert (XBYTE * source_pix, XBYTE * dest_pix, RECT & src_rect, RECT & dst_rect, RECT & src_bounds, RECT & dst_bounds, int src_pitch, int dst_pitch, int ri, int gi, int bi);

		// Clipping Functions:
		// source to destination (image/viewport) boundary clipping function:
		int ClipRects (RECT & src_rect, RECT & dst_rect, const RECT & src_bounds, const RECT & dst_bounds); // clips source and dest rects to their boundaries
		void ClipTile (int x, int y, int xres, int yres, int &mx, int &my, int &offx, int &offy);

		// helper function for functions that want to convert 4 sets of (r,g,b,a) to (r,g,b)
		void ScaleByAlpha (int & r, int & g, int & b, int & a, int & r2, int & g2, int & b2, int & a2, int & r3, int & g3, int & b3, int & a3, int & r4, int & g4, int & b4, int & a4);
	
		
		
		void (*FuncDrawPixel) (int x, int y, int r, int g, int b);
		void (*FuncDrawPixelAdded) (int x, int y, int ri, int gi, int bi);
///		int (*FuncDrawImage) (ImageX *img, RECT src_rect, RECT dst_rect);
///		int (*FuncDrawMasked) (ImageX *img, RECT src_rect, RECT dst_rect);
///		int (*FuncDrawAlpha) (ImageX *img, RECT src_rect, RECT dst_rect);
///		int (*FuncDrawAdded) (ImageX *img, RECT src_rect, RECT dst_rect, int ri, int gi, int bi);		
///		int (*FuncDrawAddedTo) (ImageX *img, RECT src_rect, RECT dst_rect, int ri, int gi, int bi, int rb, int gb, int bb);		
///		int (*FuncDrawBlended) (ImageX *img, RECT src_rect, RECT dst_rect, int ri, int gi, int bi);		
///		int (*FuncDrawMaskedBlended) (ImageX *img, RECT src_rect, RECT dst_rect, int ri, int gi, int bi);		



		void SyncWithRefresh (void); // Returns once the scanline is clear of the window


		// Windows & DirectX Specific Initialization
		bool WindowInitClass (void);								// Create a window class
		bool WindowInit (void);									// Create a main window
		bool WindowDraw (void);									// Draw window on screen
		bool DirectDrawInit (void);								// Initialize direct draw
		bool Direct3DInit (void);
		bool DirectInputInit (void);
		bool DirectSoundInit (void);
		bool DirectMusicInit (void);
		void VSyncInit (void);
		bool DirectSoundFormat (void);
		bool DirectDrawSurfInit (void);							// Initialize direct draw surfaces
		bool DirectXCleanUp (void);							// Cleanup direct draw

		// Sound Functions
		void UpdateSound (SoundX* snd);
		int AddSound (SoundX* snd);

		HINSTANCE				win_hinst;	
		int						win_hmode;

		int						win_status; 
		int						win_request_xsize, win_request_ysize; // requested width and height of view area
		int						win_request_bpp; // VIDEO_16BIT makes this 16, VIDEO_32BIT makes this 32, etc., 0 means no request
		int						win_request_full; // 0 if VIDEO_WINDOWED flag on, 1 if VIDEO_FULLSCREEN flag on, 2 if neither is on, temporarily 
		int						win_request_vsync; // 0 if VIDEO_NOREFRESHSYNC flag on, 1 if VIDEO_ALWAYSREFRESHSYNC flag on, 2 if neither on
		bool					win_request_no_flip; // false unless VIDEO_ALLOWPAGEFLIP flag off
		bool					win_request_no_sound; // false unless AUDIO_DISABLE flag on
		bool					win_request_no_stereo; // false unless AUDIO_NOSTEREO flag on
		bool					win_request_can_resize; // true unless VIDEO_NORESIZE flag on
		bool					win_request_no_stretch; // false unless VIDEO_LETTERBOX flag on
		bool					win_request_delete_mdata; // false unless RUN_ASSUMEVRAMSAFE flag on
		bool					win_request_software_only; // false unless VIDEO_SOFTWAREONLY flag on
		bool					win_request_no_auto_pause; // false unless RUN_BACKGROUND flag on
		bool					win_request_no_hide_cursor; // false unless RUN_KEEPCURSOR flag on
		bool					win_request_no_mouse_input; // false unless RUN_NOMOUSEINPUT flag on
		bool					win_request_no_manage_speed; // false unless RUN_UNMANAGED flag on
		bool					win_request_no_confirm_quit; // false unless RUN_NOCONFIRMQUIT flag on
		bool					win_request_draw_all_states; // false unless VIDEO_DRAWALLSTATES flag on
		bool					win_request_reverse_speakers; // false unless AUDIO_REVERSESPEAKERS flag on
		bool					win_request_update_between_states; // false unless VIDEO_UPDATE_BETWEEN_STATES flag on

		bool					win_dialog_up;
		bool					win_lowquality;

		bool					win_fullscreen;	
		bool					win_cursor_hidden;
		int						win_page_flipping;	
		bool					win_vsync_enabled; // true if vsync is on and currently ready
		int						win_size;
		int						win_widbits;
		int						win_scanline_speed, win_blit_latency;
//		RECT					win_destrect;

		HWND					win_hwnd;

		char					win_name [64];

		int						win_r, win_g, win_b;
		long int				win_rgb;		
		int						win_keys;
		char					win_keybuf[WINDX_BUFSIZE];
///		bool					win_keydown[256];

		int						mouse_x, mouse_y;
///		int						mouse_xl, mouse_yl;
		int						mouse_down[4];
///		int						mouse_motion;

		unsigned short *		table_alpha;

///		void *					net_variables [1000];

		float					master_music_tempo;
		float					master_music_volume;


		// vital DirectX stuff:

		// Direct3D 7.0
		LPDIRECT3D7				d3d_main;		// D3D main object
        LPDIRECT3DDEVICE7		d3d_device;		// D3D rendering device

/*		// Direct3D 8.1
		LPDIRECT3D8				d3d_main;		// D3DDevice
        LPDIRECT3DDEVICE8		d3d_device;		// Rendering device
        //D3DCAPS8 *			d3d_caps;		// Caps for the Rendering Device
		//LPD3DXSPRITE			d3d_sprite;		// Sprites
		//LPDIRECT3DTEXTURE8	d3d_texture;
		//RECT *				d3d_rect;
*/				
		int						dd_was_lost;
		HRESULT					dd_last_error;
		bool					win_no_acc_err;

		// DirectDraw 7.0		
		LPDIRECTDRAW7			dd_main;
		LPDIRECTDRAWSURFACE7	dd_front_buf;
		LPDIRECTDRAWSURFACE7	dd_back_buf;
		LPDIRECTDRAWSURFACE7	temp_surface;
		
		// DirectInput 7.0		
		LPDIRECTINPUT			di_main;
		LPDIRECTINPUTDEVICE		di_keyboard;
		LPDIRECTINPUTDEVICE		di_mouse;
		unsigned char			di_keystate[256];
		unsigned char			di_keypressed[256];
		DIMOUSESTATE			di_mousestate;
		BYTE					di_mousepressed[4];
		int						di_mousestop;

		// DirectSound 3.1
		LPDIRECTSOUND			ds_main;
		int						ds_samples_per_sec;
		int						ds_bits_per_sample;
		int						ds_num_channels;

		// DirectMusic
		IDirectMusicPerformance * dm_performance;
		IDirectMusicLoader *	  dm_loader;
		MusicX *				  dm_musicx [256];
		int						  dm_musicx_index;

		bool					sound_support;

		int						num_sounds;
		int						max_sounds;
		int						num_playing;
		SoundXPtr*				sound_list;
		LPDIRECTSOUNDBUFFER*	ds_sound_list;

		char					game_caption[200];
		char					game_name[100];
		char					gamex_ver[100];
	};

	//*** Switched Drawing Routines
	// 16-bit Drawing Routines
	static void DrawPixel16 (int x, int y, int r, int g, int b);	// Draw pixel (16-bit) -- r,g,b out of 255
	static void DrawPixelAdded16 (int x, int y, int ri, int gi, int bi);	// Draw added pixel (16-bit) -- ri,gi,bi out of 255
///	static int DrawImage16 (ImageX *img, RECT src_rect, RECT dst_rect);		// Draw image (16-bit)
///	static int DrawMasked16 (ImageX *img, RECT src_rect, RECT dst_rect);	// Draw masked image (16-bit)
///	static int DrawAlpha16 (ImageX *img, RECT src_rect, RECT dst_rect);		// Draw alpha image (16-bit)
///	static int DrawAdded16 (ImageX *img, RECT src_rect, RECT dst_rect, int ri, int gi, int bi);		// Draw added image (16-bit)
///	static int DrawAddedTo16 (ImageX *img, RECT src_rect, RECT dst_rect, int ri, int gi, int bi, int rb, int gb, int bb);		// Draw added image (16-bit)
///	static int DrawBlended16 (ImageX *img, RECT src_rect, RECT dst_rect, int ri, int gi, int bi);		// Draw blended image (16-bit)
///	static int DrawMaskedBlended16 (ImageX *img, RECT src_rect, RECT dst_rect, int ri, int gi, int bi);		// Draw masked blended image (16-bit)
	
///	static int DrawImage24 (ImageX *img, int x, int y);		// Draw image (24-bit)
///	static int DrawMasked24 (ImageX *img, int x, int y);	// Draw masked image (24-bit)
///	static int DrawAlpha24 (ImageX *img, int x, int y);		// Draw alpha image (24-bit)
	
	static HRESULT DisplayEnumCallback(LPDDSURFACEDESC2 lpDDModeDesc, LPVOID lpContext);
//		HRESULT (*DisplayEnumCallback) (LPDDSURFACEDESC2 lpDDModeDesc, LPVOID lpContext);


	#ifdef GAMEX_MAIN
		WindowsDX GameX;
		int win_maskr,	win_maskg,	win_maskb;
		int win_shiftr, win_shiftg, win_shiftb;
		int win_maxr,	win_maxg,	win_maxb;
	#else
		extern WindowsDX GameX;
		extern int win_maskr,	win_maskg,	win_maskb;
		extern int win_shiftr,	win_shiftg, win_shiftb;
		extern int win_maxr,	win_maxg,	win_maxb;		
		extern void GameInit ();
		extern void GameMain ();
		extern void GameLoop ();	// this function should update the game state but NOT draw that state
		extern void GameDraw ();	// this function should draw the game state and do as little else as possible
	#endif


#endif

